//=======================================================================
/// This program is useful generating a compressed frame
///
/// \todo
///     Support argument to specify the version of the frame
///     specification of the output frame file. 
//=======================================================================
#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

extern "C" {
#include <unistd.h>
} // extern "C"

#include <memory>
#include <sstream>
#include <string>

#include "general/AtExit.hh"
#include "general/unordered_map.hh"
#include "general/fstream.hh"

#include "framecpp/Common/Compression.hh"
#include "framecpp/Common/FrameStream.hh"

using std::string;
using General::filebuf;

using General::unordered_map;

using namespace FrameCPP::Common;

char*	Program;

string	Compression( "zero_suppress_otherwize_gzip" );
INT_4U  Level( 1 );
string	OutputName( "CompressedOut.gwf" );

typedef unordered_map< string, FrameCPP::Compression::compress_type > compression_mode_map_type;

static const compression_mode_map_type&
init_cmm( )
{
  static compression_mode_map_type	cmm;

  using namespace FrameCPP::Compression;

  //---------------------------------------------------------------------
  /// \todo This needs to be version specific
  //---------------------------------------------------------------------
  cmm[ "raw" ] = MODE_RAW;
  cmm[ "gzip" ] = MODE_GZIP;
  cmm[ "diff" ] = MODE_DIFF;
  cmm[ "diff_gzip" ] = MODE_DIFF_GZIP;
  cmm[ "zero_suppress_short" ] = MODE_ZERO_SUPPRESS_SHORT;
#if WORKING
  cmm[ "zero_suppress_short_gzip_other" ] = MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER;
#endif /* WORKING */
  cmm[ "zero_suppress_int_float" ] = MODE_ZERO_SUPPRESS_INT_FLOAT;
#if WORKING
  cmm[ "zero_suppress_otherwize_gzip" ] = MODE_ZERO_SUPPRESS_OTHERWISE_GZIP;
#endif /* WORKING */

  return cmm;

}

const compression_mode_map_type& CompressionModeMap = init_cmm( );

inline void
depart( int Status )
{
  General::AtExit::Cleanup( );
  exit( Status );
}

static void usage( )
{
  std::cerr << "Usage: "
	    << Program
	    << " [-b <size>] [-c <compression mode>] [-l <level>] [-M] [-o <output>]"
	    << " frame_file [ frame_file ... ]" << std::endl
	    << "\t-b <size> - set the size of the io buffer" << std::endl
	    << "\t-c <compression mode> - specify the compression mode" << std::endl
	    << "\t-l <level> - specify the level of compression" << std::endl
	    << "\t-M - enable memory mapped io" << std::endl
	    << "\t-o <output> - specify the output filename" << std::endl
    ;
}

int
main( int ArgC, char**ArgV )
try
{
  bool mm_io( false );
  INT_4U bs( 256 );
  INT_4U multiplier( 1024 );

  Program = ArgV[ 0 ];

  int opt;

  while( ( opt = getopt( ArgC, ArgV, "Mb:c:l:o:" ) ) != -1 )
  {
    switch( opt )
    {
    case 'b':
      {
	if ( optarg )
	{
	  INT_4U pos( strlen( optarg ) - 1);

	  switch ( optarg[ pos ] )
	  {
	  case 'k':
	    multiplier = 1024;
	    optarg[ pos ] = '\0';
	    break;
	  case 'M':
	    multiplier = 1024 * 1024;
	    optarg[ pos ] = '\0';
	    break;
	  }
	}
	std::istringstream input( optarg );
	input >> bs;
      }
      break;
    case 'c': /* compression mode */
      Compression = optarg;
      break;
    case 'l': /* compression level */
      {
	std::stringstream input( optarg );
	input >> Level;
      }
      break;
    case 'M':
      mm_io = true;
      break;
    case 'o': /* output */
      OutputName = optarg;
      break;
    default:
      break;
    } // switch( opt )
  } // while ( opt = getopt( ... )
  ArgC -= optind;
  ArgV += optind;

  if ( ArgC != 1 )
  {
    usage( );
    depart( 1 );
  } // if ( ArgC != 1 )

  //---------------------------------------------------------------------
  // Lookup the compression
  //---------------------------------------------------------------------
  INT_2U c = FrameCPP::Compression::MODE_RAW;

  
  compression_mode_map_type::const_iterator
    cm = CompressionModeMap.find( Compression );
  if ( cm != CompressionModeMap.end( ) )
  {
    c = cm->second;
  }
  else
  {
    std::ostringstream msg;
    msg << "Unknown compression mode: " << Compression;

    throw std::runtime_error( msg.str( ) );
  }

  //---------------------------------------------------------------------
  // Open the input stream
  //---------------------------------------------------------------------
  bs *= multiplier;

  General::AutoArray< CHAR > ibuffer( new CHAR[ bs ] );
  General::AutoArray< CHAR > obuffer( new CHAR[ bs ] );

  FrameBuffer< filebuf >*
    ibuf( new FrameBuffer< filebuf >( std::ios::in ) );
  FrameBuffer< filebuf >*
    obuf( new FrameBuffer< filebuf >( std::ios::out ) );

  ibuf->pubsetbuf( ibuffer.get( ), bs );
  ibuf->UseMemoryMappedIO( mm_io );

  obuf->pubsetbuf( obuffer.get( ), bs );
  obuf->UseMemoryMappedIO( mm_io );

  ibuf->open( ArgV[ 0 ], std::ios::in | std::ios::binary );
  obuf->open( OutputName.c_str( ), std::ios::out | std::ios::binary );

  IFrameStream  iframe_stream( ibuf );
  OFrameStream  oframe_stream( obuf );

  std::auto_ptr< FrameSpec::Object >
    frame( iframe_stream.ReadFrame( true /* decompress */ ) );

  while ( frame.get( ) )
  {
    //-------------------------------------------------------------------
    // Write the source frame to the output frame file
    //-------------------------------------------------------------------
    oframe_stream.WriteFrame( *frame, c, Level );
    //-------------------------------------------------------------------
    // Read the next input frame
    //-------------------------------------------------------------------
    frame = iframe_stream.ReadFrame( );
  }
  //---------------------------------------------------------------------
  // Prepare to leave
  //---------------------------------------------------------------------
  oframe_stream.Close( );
  depart( 0 );
}
catch( std::exception& e )
{
  std::cerr << "ABORT: Caught exception: " << e.what( ) << std::endl;
  depart( 1 );
}
catch( ... )
{
  std::cerr << "ABORT: Caught unknown exception: " << std::endl;
  depart( 1 );
}
